#include "camera.h"
using namespace cv;
Camera::Camera() {}

// Capture 30 frames to give autoexposure, etc. a chance to settle
void Camera::warmingUp()
{
    for (int i = 0; i < 30; ++i)
    {
        rs2::frameset data = pipe.wait_for_frames();
        auto frameDepth = data.get_depth_frame();
        //frameDepth = decFilter.process(frameDepth);
        frameDepth = spatFilter.process(frameDepth);
        frameDepth = tempFilter.process(frameDepth);
    }
}

Mat Camera::coloredFrame(Mat frameDepth)
{
    Mat depthFrameColored(frameDepth.size(), CV_8U);
    int width = frameDepth.cols, height = frameDepth.rows;
    static uint32_t histogram[0x10000];
    memset(histogram, 0, sizeof(histogram));

    for (int i = 0; i < height; ++i)
    {
        for (int j = 0; j < width; ++j)
        {
            ++histogram[frameDepth.at<ushort>(i, j)];
        }
    }

    for (int i = 2; i < 0x10000; ++i)
        histogram[i] += histogram[i - 1]; // Build a cumulative histogram for the indices in [1,0xFFFF]

    for (int i = 0; i < height; ++i)
    {
        for (int j = 0; j < width; ++j)
        {
            if (uint16_t d = frameDepth.at<ushort>(i, j))
            {
                int f = histogram[d] * 255 / histogram[0xFFFF]; // 0-255 based on histogram location
                depthFrameColored.at<uchar>(i, j) = static_cast<uchar>(f);
            }
            else
            {
                depthFrameColored.at<uchar>(i, j) = 0;
            }
        }
    }
    bitwise_not(depthFrameColored, depthFrameColored); //reverse colormap
    applyColorMap(depthFrameColored, depthFrameColored, cv::COLORMAP_JET);
    depthFrameColored.setTo(cv::Scalar(0, 0, 0), (frameDepth == 0));
    return depthFrameColored;
}

Mat Camera::captureFrame()
{
    auto frame = pipe.wait_for_frames();
    auto frameDepth = frame.get_depth_frame();
    filterDepthFrame(frameDepth);
    Mat matFrame(Size(frameDepth.get_width(), frameDepth.get_height()), CV_16UC1, (void *)frameDepth.get_data(), Mat::AUTO_STEP);
    return matFrame;
}
void Camera::captureFramesAlign()
{
    rs2::align align(RS2_STREAM_DEPTH);
    auto frameset = pipe.wait_for_frames();
    //Get processed aligned frame
    frameset = align.process(frameset);
    // Trying to get color and aligned depth frames
    rs2::video_frame color_frame = frameset.get_color_frame();
    rs2::depth_frame depth_frame = frameset.get_depth_frame();
    filterDepthFrame(depth_frame);
    matDepth = Mat(Size(depth_frame.get_width(), depth_frame.get_height()), CV_16UC1, (void *)depth_frame.get_data(), Mat::AUTO_STEP);
    Mat(Size(color_frame.get_width(), color_frame.get_height()), CV_8UC3, (void *)color_frame.get_data(), Mat::AUTO_STEP).copyTo(matRGB);
}

void Camera::filterDepthFrame(rs2::depth_frame &frameDepth)
{
    // frameDepth = decFilter.process(frameDepth);
    frameDepth = spatFilter.process(frameDepth);
    frameDepth = tempFilter.process(frameDepth);
    intrinsics = frameDepth.get_profile().as<rs2::video_stream_profile>().get_intrinsics();
}
Mat Camera::getAverageFrame(int numberFrame)
{
    Mat averageFrame;
    Camera::captureFrame().copyTo(averageFrame);
    for (int i = 1; i <= numberFrame; i++)
    {
        averageFrame *= i;
        add(averageFrame, Camera::captureFrame(), averageFrame);
        averageFrame /= i + 1;
    }
    return averageFrame;
}

Point3f Camera::deprojectPixelToPoint(float coord[], float z1)
{
    float p[3];
    rs2_deproject_pixel_to_point(p, &intrinsics, coord, z1);
    return Point3f(p[0], p[1], p[2]);
}

Point2i Camera::projectPointToPixel(Point3f point3D)
{
    float point[3] = {point3D.x, point3D.y, point3D.z};
    float pixel[2];
    rs2_project_point_to_pixel(pixel, &intrinsics, point);
    return Point2i(pixel[0], pixel[1]);
}

void Camera::start()
{
    rs2::log_to_console(RS2_LOG_SEVERITY_ERROR);
    //spatFilter.set_option(RS2_OPTION_HOLES_FILL, 5);
    profile = pipe.start(cfg);
    auto sensor = profile.get_device().first<rs2::depth_sensor>();
    // TODO: At the moment the SDK does not offer a closed enum for D400 visual presets
    // We do this to reduce the number of black pixels
    // The hardware can perform hole-filling much better and much more power efficient then our software
    auto range = sensor.get_option_range(RS2_OPTION_VISUAL_PRESET);
    for (auto i = range.min; i < range.max; i += range.step)
        if (std::string(sensor.get_option_value_description(RS2_OPTION_VISUAL_PRESET, i)) == "High Density")
            sensor.set_option(RS2_OPTION_VISUAL_PRESET, i);
    warmingUp();
}

void Camera::startAlign()
{
    rs2::log_to_console(RS2_LOG_SEVERITY_ERROR);
    spatFilter.set_option(RS2_OPTION_HOLES_FILL, 5);
    //cfg.enable_device_from_file("../input/flux/flux5.bag");
    profile = pipe.start(cfg);
    auto sensor = profile.get_device().first<rs2::depth_sensor>();
    // TODO: At the moment the SDK does not offer a closed enum for D400 visual presets
    // We do this to reduce the number of black pixels
    // The hardware can perform hole-filling much better and much more power efficient then our software
    auto range = sensor.get_option_range(RS2_OPTION_VISUAL_PRESET);
    for (auto i = range.min; i < range.max; i += range.step)
        if (std::string(sensor.get_option_value_description(RS2_OPTION_VISUAL_PRESET, i)) == "High Density")
            sensor.set_option(RS2_OPTION_VISUAL_PRESET, i);
    warmingUp();
}

void Camera::stop()
{
    pipe.stop();
}