#include "sandbox.h"
#include <numeric>
#include <fstream>
#include <string>

using namespace std;
using namespace cv;

double Sandbox::toDegrees(double radians)
{
    return radians * (180.0 / M_PI);
}

bool Sandbox::readSandboxPosition(vector<Point> &rectPoints)
{
    std::ifstream infile(SANDBOX_POSITION_FILE);
    if (infile.good())
    {
        int x, y;
        while (infile >> x >> y)
        {
            rectPoints.push_back(Point(x, y));
            cout << x << " " << y << endl;
        }
    }
    infile.close();
    return rectPoints.size() == 4;
}

void Sandbox::writeSandboxPosition(vector<Point> rectPoints)
{
    //export points to file
    std::ofstream myfile;
    myfile.open(SANDBOX_POSITION_FILE);
    for (unsigned int i = 0; i < rectPoints.size(); i++)
        myfile << rectPoints[i].x << " " << rectPoints[i].y << endl;
    myfile.close();
}

void Sandbox::sanboxBorder()
{    
    char windowName[] = "border";
    createWindowsFullScreen(windowName);
    Mat frameBeamer(Size(Beamer::width, Beamer::height), CV_8UC3, Scalar(255, 0, 0));
    
    // Show blue frame
    imshow(windowName, frameBeamer);
    waitKey(100);

    camera.startAlign(); // 1 seconde of warming up
    camera.captureFramesAlign();

    Mat frameData = camera.getDepthFrameAlign();
    Mat coloredFrame = camera.getRGBFrameAlign();
    Size s = frameData.size();
     
    // calibrate distance
    Point center(s.width / 2, s.height / 2);
    float distancePlan = static_cast<float>(mean(frameData(Rect(center, Size(10, 10))))[0]); // get sandbox distance from center of camera
    calibrate.setDistancePlan(distancePlan);
    destroyAllWindows();

    vector<Point> rectPoints;
    float y = coloredFrame.size().height;
    float x = coloredFrame.size().width;
    rectPoints.push_back(Point(1.0/4*x, 1.0/4*y));
    rectPoints.push_back(Point(1.0/4*x, 3.0/4*y));
    rectPoints.push_back(Point(3.0/4*x, 3.0/4*y));
    rectPoints.push_back(Point(3.0/4*x, 1.0/4*y));

    cout << "Edit Rectangle" << endl;
    BorderEdit::edit(coloredFrame, &rectPoints); // edit projected frame

    // adjust model matrixe to forme a rectangle in sandbox
    int widthTop = rectPoints[3].x - rectPoints[0].x;
    double angle1 = atan((double)(rectPoints[3].y - rectPoints[0].y) / widthTop);
    Mat matRotation = getRotationMatrix2D(center, toDegrees(angle1), 1);
    calibrate.setMatrixRotation(matRotation);
    cout << "Calibrate Rotation Matrixe" << endl;

    Size rectSize = Size(widthTop, cvRound(widthTop / 1.33333) + 5);
    Point p = calibrate.rotatePixel(rectPoints[0]);
    rectSandbox = Rect(p, rectSize);
    writeSandboxPosition(rectPoints); // update with adjusted sandbox's angles
    destroyAllWindows();

    /* 
    
        TODO : save distancePlan in config_file
        
    */
    cout << "Saving distance in file..." << endl;
    std::ofstream myfile;
    myfile.open(CALIBRATE_DISTANCE_FILE);
    myfile << distancePlan << endl;
    myfile.close();
    cout << "Done!" << endl;
    /* 
    
        TODO : save matRotation in config_file
        
    */
    cout << "Saving matrixe in file..." << endl;
    myfile.open(CALIBRATE_MATRIX_FILE);
    cout << "Data : " << matRotation.rows << " : " << matRotation.cols << endl;
    myfile << matRotation.cols << " " << matRotation.rows << endl;
    for (int y = 0; y < matRotation.rows; y++)
    {
        for (int x = 0; x < matRotation.cols; x++)
        {
            myfile << matRotation.at<uint16_t>(y, x) << CHAR_DELIM;
        }
        myfile << endl;
    }
    myfile.close();
    cout << "Done..." << endl;
}

void Sandbox::sandboxBorderLoad(){
    float distancePlan;
    Mat matRotation;

    // Distance Plan
    std::ifstream infile(CALIBRATE_DISTANCE_FILE);
    if (infile.good()){
        while (infile >> distancePlan){
            cout << distancePlan << endl;
        }
    }
    infile.close();

    // Matrix Rotation
    char *filename = "";
    strcpy(filename, CALIBRATE_MATRIX_FILE);
    importFrame(filename, matRotation);


    calibrate.setDistancePlan(distancePlan);
}

void Sandbox::exportToAsc(Mat src, char *fileName)
{
    std::ofstream myfile;
    myfile.open(fileName);
    Mat frameSand = src(rectSandbox);
    for (int y = 0; y < src.rows; y++)
    {
        for (int x = 0; x < src.cols; x++)
        {
            myfile << x << " " << y << " " << frameSand.at<uint16_t>(y, x) << "\n";
        }
    }
    myfile.close();
}

void Sandbox::exportFrame(Mat src, char *fileName)
{
    std::ofstream myfile;
    myfile.open(fileName);
    Mat frameSand = src(rectSandbox);
    myfile << frameSand.cols << " " << frameSand.rows << endl;
    for (int y = 0; y < frameSand.rows; y++)
    {
        for (int x = 0; x < frameSand.cols; x++)
        {
            myfile << frameSand.at<uint16_t>(y, x) << CHAR_DELIM;
        }
        myfile << endl;
    }
    myfile.close();
}

bool Sandbox::importFrame(char *fileName, Mat &output)
{

    std::ifstream file(fileName);
    if (file.is_open())
    {
        string elem;
        int i = 0;
        //read header
        getline(file, elem, CHAR_DELIM);
        int width = atoi(elem.c_str());
        getline(file, elem, CHAR_DELIM);
        int height = atoi(elem.c_str());
        //create & read matrix
        output = Mat(Size(width, height), CV_16UC1);
        cout << width << " " << height << endl;
        while (getline(file, elem, CHAR_DELIM))
        {
            output.at<uint16_t>(i / width, i % width) = (uint16_t)atoi(elem.c_str());
            i++;
        }
        file.close();
        return true;
    }
    return false;
}

bool Sandbox::configFilesFound(){
    const char* files[] = { SANDBOX_POSITION_FILE,
                            CALIBRATE_DISTANCE_FILE,
                            CALIBRATE_MATRIX_FILE };
    int size = (sizeof(files)/sizeof(char*));
    
    for (int i = 0; i < size; i++)
    {
        char *filename = "";
        strcpy(filename, files[i]);
        std::ifstream infile(filename);
        if(infile.fail()){
            cout << "Configuration file missing...";
            return false;
        }
    }

    return true;
}

void Sandbox::createWindowsFullScreen(char *windowName)
{
    namedWindow(windowName, CV_WINDOW_NORMAL);
    setWindowProperty(windowName, CV_WND_PROP_FULLSCREEN, CV_WINDOW_FULLSCREEN);
}

void Sandbox::showImage(Mat image, char *windowName)
{
    static Mat frameBeamer(Size(Beamer::width, Beamer::height), CV_8UC3);
    camera.captureFramesAlign();
    Mat depth = camera.getDepthFrameAlign()(rectSandbox);
    resize(image, image, depth.size());
    Mat imageCalibrate(depth.size(), CV_8UC3, Scalar(0, 0, 0));
    calibrate.transformationFrame(depth, image, imageCalibrate, camera, beamer.getPosition());
    //flip to align frame with beamer
    flip(imageCalibrate, imageCalibrate, 1);
    flip(imageCalibrate, imageCalibrate, 0);
    resize(imageCalibrate, frameBeamer, frameBeamer.size());
    imshow(windowName, frameBeamer);
}

void Sandbox::showImage(cv::Mat image)
{
    showImage(image, defaultWindowsName);
}
Mat Sandbox::getRGBFrame()
{
    camera.captureFramesAlign();
    return camera.getRGBFrameAlign()(rectSandbox);
}
Mat Sandbox::getDepthFrame()
{
    camera.captureFramesAlign();
    return camera.getDepthFrameAlign()(rectSandbox);
}
Sandbox::Sandbox()
{
    bool configSanbox = true;
    //cout << "All files there? " << configFilesFound() << endl; 
    if(configSanbox || !configFilesFound()){
        // Blue screen and edit colored frame routine
        sanboxBorder();
        // Matching cross under beamer routine
        beamer.findBeamer(camera);
    }else{
        sandboxBorderLoad();
        //beamer.findBeamerLoad(camera);
    }
    // Set full screen parameter for show function
    createWindowsFullScreen(defaultWindowsName);
}
